#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2012-2015  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# (version 2) as published by the Free Software Foundation.
#
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
# TODO
* add valign parameters (or attribute)
"""

import itertools
import json
import re
import xml.dom.minidom

from types import GeneratorType

# row, widths, alignments
def tzip(fields, widths, alignments, default_alignment='l'):
  """
  A zipper function for generating tuples of fields, widths and alignments.

  fields: The list or tuple of field elements.

  widths: The widths corresponding to each field.

  alignments: The alignments for each field. If empty, the default alignment
  will be used. If not empty but shorter than the number of fields and widths,
  the last given alignment will be used for the missing values. Valid
  alignments: 'l' for left, 'c' for center, 'r' for right

  default_alignment: The default alignment to use if the alignments are empty.
  """
  for i in range(len(widths)):
    try:
      field = fields[i]
    except IndexError:
      field = None
    try:
      yield (field, widths[i], alignments[i])
      default_alignment = alignments[i]
    except IndexError:
      yield (field, widths[i], default_alignment)


def pad(string, width, alignment, char=' ', len_func=None):
  """
  Pad a string to a given width with a given alignment and padding character.

  string: the string

  width: the width

  alignment: the alignment ('l': left, 'c': center, 'r': right)

  char: the character to ue for padding (default: ' ')

  len_func: A function for calculating the length. Use this when the output will
  contain invisible characters such as ANSI escapes or other characters that
  should be ignored for column width calculations.
  """
#   string = string.replace('\n', ' ')
#   string = string.replace('\t', ' ')
  if string is None:
    string = ''
  z = []
  for y in string.split('\n'):
    if len_func:
      dw = len(y) - len_func(y)
    else:
      dw = 0
    if alignment == 'l':
      y = y.ljust(width + dw, char)
    elif alignment == 'c':
      y = y.center(width + dw, char)
    else:
      y = y.rjust(width + dw, char)
    z.append(y)
  return '\n'.join(z)



def homogenize(rows, cols=None, pad=' ', align='l', len_func=len):
  """
  Pad all cells in a given column to the same width.

  cols: indices of columns to homogenize. All columns will be aligned by
  default.

  pad: character to use for padding.

  align: alignment direction. l: left, r:right, c:center

  len_func: A function for calculating the length. Use this when the output will
  contain invisible characters such as ANSI escapes or other characters that
  should be ignored for column width calculations.

  This is useful for centering a column in the table body with a header that is
  larger than any other cellin the table body. For example, a table of element
  symbols and weights would have a column of element symbols no wider than 3
  characters, but the header might be "Symbol". This function could then be used
  to center the symbols in the column while still aligning the left character of
  each symbol.
  """
  if isinstance(rows, GeneratorType):
    rows = tuple(rows)
  if None == cols:
    cols = tuple(range(max(len(r) for r in rows)))
  widths = dict.fromkeys(cols, 0)
  for row in rows:
    if isinstance(row, GeneratorType):
      row = tuple(row)
    for i in cols:
      try:
        x = row[i]
        widths[i] = max(widths[i], len_func(x))
      except IndexError:
        pass
  for row in rows:
    row = list(row)
    for i in cols:
      try:
        x = row[i]
        if align == 'r':
          x = x.rjust(widths[i])
        elif align == 'c':
          x = x.center(widths[i])
        else:
          x = x.ljust(widths[i])
        row[i] = x
      except IndexError:
        pass
    yield row



def align_char(rows, cols=None, char='.', pad=' ', len_func=len):
  """
  Align elements of each column along a given character (or string). This is
  useful for e.g. aligning columns along decimal points.

  cols: indices of columns to align. All columns will be aligned by default.

  char: the character (or string) by which to align. Default: "."

  pad: the character to use for padding. Default: " "

  len_func: A function for calculating the length. Use this when the output will
  contain invisible characters such as ANSI escapes or other characters that
  should be ignored for column width calculations.
  """
  if isinstance(rows, GeneratorType):
    rows = tuple(rows)
  if None == cols:
    cols = tuple(range(max(len(r) for r in rows)))
  l = len(char) # just in case it's not a single character
  left = dict.fromkeys(cols, 0)
  right = dict.fromkeys(cols, 0)
  for row in rows:
    if isinstance(row, GeneratorType):
      row = tuple(row)
    for i in cols:
      try:
        x = row[i]
        if char in x:
          a,b = x.split(char, 1)
          left[i] = max(left[i], len_func(a))
          # +1 because the character will count towards the right "column"
          right[i] = max(right[i], len_func(b))
        else:
          left[i] = max(left[i], len_func(x))
      except IndexError:
        pass
  for row in rows:
    r = list(row)
    for i in cols:
      try:
        x = r[i]
        if char in x:
          a,b = x.split(char, 1)
          a = a.rjust(left[i], pad)
          b = b.ljust(right[i], pad)
          r[i] = char.join((a,b))
        else:
          # If the right "column" is empty then just pad the left.
          if right[i] == 0:
            r[i] = x.rjust(left[i])
          # Else add extra on the right for the char.
          else:
            r[i] = x.rjust(left[i]) + pad * (right[i] + l)
      except IndexError:
        pass
    yield r



class Tabulator():
  """
  Arrange data into tables.
  """

  table_types = ('plain', 'xhtml', 'box', 'markdown', 'latex', 'json')

  def __init__(self, rows, cols=None, headers=None, len_func=len):
    """
    rows: An iterable of rows, where each row itself is an iterable.

    cols: An iterable of column alignments.

    headers: An iterable of column alignments. If given, the first row will be
    treated as headers with the formatting of this column applied.

    len_func: A function for calculating the length. Use this when the output
    will contain invisible characters such as ANSI escapes or other characters
    that should be ignored for column width calculations.

    The alignment iterables are sequences of the characters "l", "c" and "r",
    specifying "left", "center" and "right" alignment in the column,
    respectively. The default is 'l'. If there are not enough entries to align
    all values in the row then the last entry will be repeated, so ('l', 'r')
    would left-align the first column, then right-align all columns after it.

    Some output formats support 'j' ("justify") as well.

    Cell data must already be in string format.
    """
    self.len_func = len_func
    self.rows = rows
    self.cols = cols
    self.headers = headers


  def len(self, s):
    return self.len_func(s)


  def __getattr__(self, attr):
    # Return tuples of tuples to prevent in-place changes to the list.
    if attr == 'rows':
      return tuple(map(tuple, self.__dict__[attr]))



  def __setattr__(self, attr, value):
    if attr == 'rows':
      widths = []
      rows = list()
      for r in value:
        if isinstance(r, GeneratorType):
          r = tuple(r)
        row = []
        for i in range(len(r)):
          cell = r[i]
          if not isinstance(cell, str):
            cell = str(cell)
          if '\n' in cell:
            w = max(self.len(x) for x in cell.split('\n'))
          else:
            w = self.len(cell)
          row.append(cell)
          try:
            widths[i] = max(widths[i], w)
          except IndexError:
            widths.append(w)
        rows.append(row)
      self.__dict__['widths'] = widths
      self.__dict__[attr] = rows

    elif attr in ('cols', 'headers'):
      if not value:
        value = tuple()
      self.__dict__[attr] = value

    elif attr == 'widths':
      raise AttributeError("'widths' is read-only")

    else:
      self.__dict__[attr] = value



  def get_rows(self):
    """Get an iterator over the rows of formatted cells."""
    rows = self.rows
    if self.headers:
      yield (pad(h,w,a,len_func=self.len) for h,w,a in tzip(rows[0], self.widths, self.headers))
      rows = rows[1:]
    for row in rows:
      yield (pad(h,w,a,len_func=self.len) for h,w,a in tzip(row, self.widths, self.cols))



  def __str__(self):
    return self.to_plain_table()



  def to_table(self, typ='plain'):
    '''
    A simple wrapper around the to_*_table functions.
    '''
    if typ == 'xhtml':
      return self.to_xhtml_table()
    elif typ == 'box':
      return self.to_box_table()
    elif typ == 'markdown':
      return self.to_markdown_table()
    elif typ == 'latex':
      return self.to_latex_table()
    elif typ == 'json':
      return self.to_json()
    else:
      return self.to_plain_table()



  def to_plain_table(self, spacer='\t'):
    table = ''
    fmt = spacer.join('{}' for w in self.widths) + '\n'
    rules = fmt.format(*tuple(pad('',w,tuple(),'-',len_func=self.len) for w in self.widths))
    added_rules = False
    for row in self.get_rows():
#       table += fmt % tuple(row)
#       if self.headers and not added_rules:
#         table += rules
#         added_rules = True
#     return table
      for line in itertools.zip_longest(*(x.split('\n') for x in row)):
        for i in range(len(line)):
          if line[i] == None:
            line = list(line)
            line[i] = ' ' * self.widths[i]
        table += fmt.format(*line)
#       table += fmt % tuple(row)
      if self.headers and not added_rules:
        table += rules
        added_rules = True
    return table



  def to_xhtml_table(self, indent='  ', pre=False):
    """
    Create an XHTML representation of the table.

    indent: intdnt argument to pass to toprettyxml

    pre: wrap cells with "pre" tags

    Only the table is returned, not a complete XHTML page.
    """
    doc = xml.dom.minidom.Document()
    table = doc.createElement('table')
    doc.appendChild(table)

    tbody = doc.createElement('tbody')
    table.appendChild(tbody)

    align = {
      'l' : 'left',
      'c' : 'center',
      'r' : 'right',
      'j' : 'justify',
    }

    added_headers = False
    for row in self.rows:
      if self.headers and not added_headers:
        thead = doc.createElement('thead')
        table.insertBefore(thead, tbody)
        tr = doc.createElement('tr')
        thead.appendChild(tr)
        for r, w, a in tzip(row, self.widths, self.headers):
          th = doc.createElement('th')
          th.setAttribute('align', align[a])
          tr.appendChild(th)
          text = doc.createTextNode(r)
          if pre:
            p = doc.createElement('pre')
            th.appendChild(p)
            p.appendChild(text)
          else:
            th.appendChild(text)
        added_headers = True
      else:
        tr = doc.createElement('tr')
        tbody.appendChild(tr)
        for r, w, a in tzip(row, self.widths, self.cols):
          td = doc.createElement('td')
          td.setAttribute('align', align[a])
          tr.appendChild(td)
          text = doc.createTextNode(r)
          if pre:
            p = doc.createElement('pre')
            td.appendChild(p)
            p.appendChild(text)
          else:
            td.appendChild(text)
        added_headers = True

    x = table.toprettyxml(indent=indent)
    # Correct the textnode output bug.
    x = re.sub(r'>\s*\n\s*([^\s<])', r'>\1', x)
    x = re.sub(r'([^\s>])\s*\n\s*<', r'\1<', x)
    x = re.sub(r'>\s*\n\s*\n\s*<', '><', x)
    return x



  def to_box_table(self, double=False, top=False, bottom=False, middle=False, left=False, right=False, center=False, between_rows=False):
    """
    Create a table drawn with UTF-8 box-drawing characters.

    "double" sets the secondary style to double lines. If false, thick lines
    will be used.

    "top", "bottom", etc. turn secondary style on or off for that border.
    """

    # Yeah yeah, I know, but hey, I was tired and curious.
    tls = (0x250c, 0x250d, 0x250e, 0x250f, 0x250c, 0x2552, 0x2553, 0x2554)
    trs = tuple(x + 4 for x in tls[:5]) + tuple(x + 3 for x in tls[5:])
    bls = tuple(x + 4 for x in trs[:5]) + tuple(x + 3 for x in trs[5:])
    brs = tuple(x + 4 for x in bls[:5]) + tuple(x + 3 for x in bls[5:])
    mls = (0x251c, 0x251d, 0x2520, 0x2523, 0x251c, 0x255e, 0x255f, 0x2560)
    mrs = tuple(x + 8 for x in mls[:5]) + tuple(x + 3 for x in mls[5:])
    tcs = (0x252c, 0x252f, 0x2530, 0x2533, 0x252c) + tuple(x + 3 for x in mrs[5:])
    bcs = tuple(x + 8 for x in tcs[:5]) + tuple(x + 3 for x in tcs[5:])
    mcs = (0x253c, 0x253f, 0x2542, 0x254b, 0x253c) + tuple(x + 3 for x in bcs[5:])
    ts = (0x2500, 0x2501, 0x2500, 0x2550)
    ls = (0x2502, 0x2503, 0x2502, 0x2551)

    top = 1 if top else 0
    bottom = 1 if bottom else 0
    middle = 1 if middle else 0
    left = 2 if left else 0
    right = 2 if right else 0
    center = 2 if center else 0

    double = 4 if double else 0

    tl = chr(tls[top + left + double])
    tr = chr(trs[top + right + double])
    bl = chr(bls[bottom + left + double])
    br = chr(brs[bottom + right + double])
    ml = chr(mls[middle + left + double])
    mr = chr(mrs[middle + right + double])
    tc = chr(tcs[top + center + double])
    bc = chr(bcs[bottom + center + double])
    mc = chr(mcs[middle + center + double])

    double = int(double / 2)
    left = int(left / 2)
    right = int(right / 2)
    center = int(center / 2)

    t = chr(ts[top + double])
    b = chr(ts[bottom + double])
    m = chr(ts[middle + double])
    l = chr(ls[left + double])
    r = chr(ls[right + double])
    c = chr(ls[center + double])

    top    = tl + tc.join(t * w for w in self.widths) + tr + '\n'
    middle = ml + mc.join(m * w for w in self.widths) + mr + '\n'
    bottom = bl + bc.join(b * w for w in self.widths) + br + '\n'
    fmt    = l  + c.join('{}' for w in self.widths)   + r  + '\n'

    table = top
    added_rules = False
    add_row_sep = False
    for row in self.get_rows():
      if add_row_sep:
        table += middle
        add_row_sep = False
      for line in itertools.zip_longest(*[x.split('\n') for x in row]):
        for i in range(len(line)):
          if line[i] == None:
            line = list(line)
            line[i] = ' ' * self.widths[i]
        table += fmt.format(*line)
      if self.headers and not added_rules:
        added_rules = True
        add_row_sep = True
      elif between_rows:
        add_row_sep = True
    table += bottom
    return table



  def to_markdown_table(self):
    """
    Create a markdown grid table as recognized by Pandoc.
    """

    top    = '+' + '+'.join('-' * w for w in self.widths) + '+\n'
    middle = '+' + '+'.join('=' * w for w in self.widths) + '+\n'
    fmt    = '|'  + '|'.join('{}' for w in self.widths)   + '|\n'

    table = top
    added_rules = False
    for row in self.get_rows():
#       row = tuple(row)
      for line in itertools.zip_longest(*[x.split('\n') for x in row]):
        for i in range(len(line)):
          if line[i] == None:
            line = list(line)
            line[i] = ' ' * self.widths[i]
        table += fmt.format(*line)
      if self.headers and not added_rules:
        table += middle
        added_rules = True
      else:
        table += top
    return table


  # TODO
  # Implement this. Proper escapes may be tricky.
  def to_latex_table(
    self,
    top='',
    middle=r'\hline',
    end=r'\hline',
    indent='  ',
  ):
    """
    Create a LaTex representation of the table.

    The current implementation is still quite primitive.
    """
    multiline_cell_fmt = r'\begin{{tabular}}[t]{{@{{}}%c@{{}}}{}\end{{tabular}}'

    n = max(len(tuple(r)) for r in self.get_rows())
    cols = list()
    last_col = 'l'
    for i in range(n):
      try:
        last_col = self.cols[i]
      except IndexError:
        pass
      cols.append(last_col)

    added_rules = False
    table = ''
    for row in self.get_rows():
      table += indent + ' & '.join(
        f.replace('&', r'\&') if '\n' not in f \
        else multiline_cell_fmt.format(a, f.replace('&', r'\&').replace('\n', r'\\')) \
        for a,f in zip(cols,row)
      )
      table += r'\\' + '\n'
      if self.headers and not added_rules:
        table += indent + middle + '\n'
        added_rules = True

    if end:
      table += indent + end
    else:
      table = table[:-3]

    return r'''\begin{{tabular}}{{{}}}
{}
\end{{tabular}}
'''.format(' '.join(cols), table)



  def to_json(self, strip=False, simple=True, sort_keys=True, indent=None):
    """
    Return the table in JSON format.

    strip: If True, entries will be stripped of leading and trailing whitespace.

    simple: If True, the rows will be a list of lists. If False, each row will
    be dictionary with the headers used as field names. This obviously requires
    headers.
    """
    if self.headers:
      if strip:
        headers = [h.strip() for h in self.rows[0]]
        # Deliberately left as a generator.
        rows = ((f.strip() for f in r) for r in self.rows[1:])
      else:
        headers = self.rows[0]
        rows = self.rows[1:]

      if simple:
        obj = dict()
        obj['headers'] = headers
        obj['rows'] = [list(r) for r in rows]
      else:
        obj = [dict(zip(self.rows[0], r)) for r in self.rows[1:]]
    else:
      if strip:
        obj = [list(f.strip() for f in r) for r in self.rows]
      else:
        obj = self.rows

    return json.dumps(obj, sort_keys=sort_keys, indent=indent)

